/*
 * Copyright (c) 2023–2024 DTB <trinity@trinity.moe>
 * Copyright (c) 2024 Emma Tebibyte <emma@tebibyte.media>
 * SPDX-License-Identifier: AGPL-3.0-or-later
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see https://www.gnu.org/licenses/.
 */

#include <stdio.h> /* fprintf(3), stderr, NULL */
#include <stdlib.h> /* EXIT_FAILURE, EXIT_SUCCESS */
#include <string.h> /* memset(3), strchr(3) */
#include <sysexits.h> /* EX_OSERR, EX_USAGE */
#include <unistd.h> /* access(3), getopt(3), pledge(2), unveil(2), F_OK, R_OK,
					 * W_OK, X_OK */
#include <sys/stat.h> /* lstat(3), stat struct, S_ISBLK, S_ISCHR, S_ISDIR,
                       * S_ISFIFO, S_ISGID, S_ISREG, S_ISLNK, S_ISSOCK,
                       * S_ISUID, S_ISVTX */

char *program_name = "scrut";
#define OPTS "bcdefgkprsuwxLS"
static char opts[] = OPTS;

static int
usage(char *argv0) {
	(void)fprintf(stderr, "Usage: %s [-" OPTS "] file...\n", argv0);

	return EX_USAGE;
}

int main(int argc, char *argv[]) {
	char sel[(sizeof opts) / (sizeof *opts)];

	program_name = argv[0] == NULL ? program_name : argv[0];

#ifdef __OpenBSD__
	if (pledge("rpath stdio unveil", NULL) == -1) {
		perror(program_name);
		return EX_OSERR;
	}
#endif

	if (argc < 2) { return usage(program_name); }

	{ /* option parsing */
		char *p;

		memset(sel, '\0', sizeof sel);
		for (int c; (c = getopt(argc, argv, opts)) != -1;) {
			if ((p = strchr(opts, c)) == NULL) { return usage(argv[0]); }
			else { sel[p - opts] = c; }
		}

		/* straighten out selections; permute out nulls */
		p = sel;
		for (size_t i = 0; i < (sizeof sel) / (sizeof *sel); ++i) {
			if (sel[i] != '\0') {
				*p = sel[i];
				if (&sel[i] != p++) { sel[i] = '\0'; }
			}
		}
	}

	if (optind == argc) { return usage(argv[0]); }

	for (argv += optind ; *argv != NULL; ++argv) {
		struct stat buf;

#ifdef __OpenBSD__
		if (unveil(*argv, "r") == -1) {
			perror(program_name);
			return EX_OSERR;
		}
#endif

		if(access(*argv, F_OK) != 0 || lstat(*argv, &buf) == -1) {
			return EXIT_FAILURE; /* doesn't exist or isn't stattable */
		}

		for (size_t i = 0; sel[i] != '\0'; ++i) {
			if (
					(sel[i] == 'b' && !S_ISBLK(buf.st_mode))
					|| (sel[i] == 'c' && !S_ISCHR(buf.st_mode))
					|| (sel[i] == 'd' && !S_ISDIR(buf.st_mode))
					|| (sel[i] == 'e' && 0)
					|| (sel[i] == 'f' && !S_ISREG(buf.st_mode))
					|| (sel[i] == 'g' && !(buf.st_mode & S_ISGID))
					|| (sel[i] == 'k' && !(buf.st_mode & S_ISVTX))
					|| (sel[i] == 'p' && !S_ISFIFO(buf.st_mode))
					|| (sel[i] == 'r' && access(*argv, R_OK) != 0)
					|| (sel[i] == 'u' && !(buf.st_mode & S_ISUID))
					|| (sel[i] == 'w' && access(*argv, W_OK) != 0)
					|| (sel[i] == 'x' && access(*argv, X_OK) != 0)
					|| (sel[i] == 'L' && !S_ISLNK(buf.st_mode))
					|| (sel[i] == 'S' && !S_ISSOCK(buf.st_mode))
			) { return EXIT_FAILURE; }
		}
	}

	return EXIT_SUCCESS;
}
